import fetch from 'node-fetch';
import { RequestInit } from 'node-fetch';
import { ClassValidator, ValidationError } from './common';
import { Document, Query, ClientContext, UpdateOptions, Version, Patch, VersionedMaybeObject, PollOptions } from './interfaces';
export interface Options extends Pick<RequestInit, 'agent' | 'redirect' | 'follow' | 'compress'> {
    fetchImplementation?: typeof fetch;
    timeoutMs?: number;
    headers?: Record<string, string>;
}
export declare class RequestError extends Error {
    /**
     * The original error causing this request to fail
     * Inherits Error in case of network or parse errors
     * In case of an invalid HTTP response it will contain an object with the body/trimmed text of the response
     */
    readonly cause: any;
    readonly method: string;
    readonly options: any;
    readonly name = "RequestError";
    constructor(message: string, 
    /**
     * The original error causing this request to fail
     * Inherits Error in case of network or parse errors
     * In case of an invalid HTTP response it will contain an object with the body/trimmed text of the response
     */
    cause: any, method: string, options: any);
}
export declare class TimeoutError extends Error {
    readonly method: string;
    readonly options: any;
    readonly name = "TimeoutError";
    constructor(message: string, method: string, options: any);
}
export { ValidationError, };
export declare type Context = ClientContext;
export interface DB {
    get(ctx: ClientContext, key: string): Promise<any>;
    getWithVersion(ctx: ClientContext, key: string): Promise<VersionedMaybeObject>;
    setIfVersion(ctx: ClientContext, key: string, version: Version, value?: ({}) | (Array<any>) | ((null) | (string) | (number) | (boolean)), opts?: UpdateOptions): Promise<boolean>;
    startPolling(ctx: ClientContext, key: string): Promise<VersionedMaybeObject>;
    poll(ctx: ClientContext, keysToVersions: Array<[string, Version]>, opts: PollOptions): Promise<Array<[string, Array<Patch>]>>;
    create(ctx: ClientContext, key: string, value: ({}) | (Array<any>) | ((null) | (string) | (number) | (boolean))): Promise<boolean>;
    remove(ctx: ClientContext, key: string): Promise<boolean>;
    find(ctx: ClientContext, query: Query): Promise<Array<Document>>;
}
export declare class DBClient {
    readonly serverUrl: string;
    protected readonly options: Options;
    static readonly methods: string[];
    static readonly validators: ClassValidator;
    protected readonly props: {
        "get": {
            "description": string;
            "type": string;
            "properties": {
                "params": {
                    "type": string;
                    "properties": {
                        "key": {
                            "type": string;
                        };
                    };
                    "propertyOrder": string[];
                    "required": string[];
                };
                "returns": {};
            };
            "propertyOrder": string[];
            "required": string[];
        };
        "getWithVersion": {
            "description": string;
            "type": string;
            "properties": {
                "params": {
                    "type": string;
                    "properties": {
                        "key": {
                            "type": string;
                        };
                    };
                    "propertyOrder": string[];
                    "required": string[];
                };
                "returns": {
                    "$ref": string;
                };
            };
            "propertyOrder": string[];
            "required": string[];
        };
        "setIfVersion": {
            "description": string;
            "type": string;
            "properties": {
                "params": {
                    "type": string;
                    "properties": {
                        "key": {
                            "type": string;
                        };
                        "version": {
                            "$ref": string;
                        };
                        "value": {
                            "anyOf": ({
                                "type": string;
                                "properties": {};
                                "items"?: undefined;
                            } | {
                                "type": string;
                                "items": {};
                                "properties"?: undefined;
                            } | {
                                "type": string[];
                                "properties"?: undefined;
                                "items"?: undefined;
                            })[];
                        };
                        "opts": {
                            "$ref": string;
                        };
                    };
                    "propertyOrder": string[];
                    "required": string[];
                };
                "returns": {
                    "type": string;
                };
            };
            "propertyOrder": string[];
            "required": string[];
        };
        "startPolling": {
            "description": string;
            "type": string;
            "properties": {
                "params": {
                    "type": string;
                    "properties": {
                        "key": {
                            "type": string;
                        };
                    };
                    "propertyOrder": string[];
                    "required": string[];
                };
                "returns": {
                    "$ref": string;
                };
            };
            "propertyOrder": string[];
            "required": string[];
        };
        "poll": {
            "description": string;
            "type": string;
            "properties": {
                "params": {
                    "type": string;
                    "properties": {
                        "keysToVersions": {
                            "type": string;
                            "items": {
                                "type": string;
                                "items": ({
                                    "type": string;
                                    "$ref"?: undefined;
                                } | {
                                    "$ref": string;
                                    "type"?: undefined;
                                })[];
                                "minItems": number;
                                "additionalItems": {
                                    "anyOf": ({
                                        "type": string;
                                        "$ref"?: undefined;
                                    } | {
                                        "$ref": string;
                                        "type"?: undefined;
                                    })[];
                                };
                            };
                        };
                        "opts": {
                            "$ref": string;
                        };
                    };
                    "propertyOrder": string[];
                    "required": string[];
                };
                "returns": {
                    "type": string;
                    "items": {
                        "type": string;
                        "items": ({
                            "type": string;
                            "items"?: undefined;
                        } | {
                            "type": string;
                            "items": {
                                "$ref": string;
                            };
                        })[];
                        "minItems": number;
                        "additionalItems": {
                            "anyOf": ({
                                "type": string;
                                "items"?: undefined;
                            } | {
                                "type": string;
                                "items": {
                                    "$ref": string;
                                };
                            })[];
                        };
                    };
                };
            };
            "propertyOrder": string[];
            "required": string[];
        };
        "create": {
            "description": string;
            "type": string;
            "properties": {
                "params": {
                    "type": string;
                    "properties": {
                        "key": {
                            "type": string;
                        };
                        "value": {
                            "anyOf": ({
                                "type": string;
                                "properties": {};
                                "items"?: undefined;
                            } | {
                                "type": string;
                                "items": {};
                                "properties"?: undefined;
                            } | {
                                "type": string[];
                                "properties"?: undefined;
                                "items"?: undefined;
                            })[];
                        };
                    };
                    "propertyOrder": string[];
                    "required": string[];
                };
                "returns": {
                    "type": string;
                };
            };
            "propertyOrder": string[];
            "required": string[];
        };
        "remove": {
            "description": string;
            "type": string;
            "properties": {
                "params": {
                    "type": string;
                    "properties": {
                        "key": {
                            "type": string;
                        };
                    };
                    "propertyOrder": string[];
                    "required": string[];
                };
                "returns": {
                    "type": string;
                };
            };
            "propertyOrder": string[];
            "required": string[];
        };
        "find": {
            "description": string;
            "type": string;
            "properties": {
                "params": {
                    "type": string;
                    "properties": {
                        "query": {
                            "$ref": string;
                        };
                    };
                    "propertyOrder": string[];
                    "required": string[];
                };
                "returns": {
                    "type": string;
                    "items": {
                        "$ref": string;
                    };
                };
            };
            "propertyOrder": string[];
            "required": string[];
        };
    };
    readonly validators: ClassValidator;
    constructor(serverUrl: string, options?: Options);
    get(ctx: ClientContext, key: string, options?: Options): Promise<any>;
    getWithVersion(ctx: ClientContext, key: string, options?: Options): Promise<VersionedMaybeObject>;
    setIfVersion(ctx: ClientContext, key: string, version: Version, value?: ({}) | (Array<any>) | ((null) | (string) | (number) | (boolean)), opts?: UpdateOptions, options?: Options): Promise<boolean>;
    startPolling(ctx: ClientContext, key: string, options?: Options): Promise<VersionedMaybeObject>;
    poll(ctx: ClientContext, keysToVersions: Array<[string, Version]>, opts: PollOptions, options?: Options): Promise<Array<[string, Array<Patch>]>>;
    create(ctx: ClientContext, key: string, value: ({}) | (Array<any>) | ((null) | (string) | (number) | (boolean)), options?: Options): Promise<boolean>;
    remove(ctx: ClientContext, key: string, options?: Options): Promise<boolean>;
    find(ctx: ClientContext, query: Query, options?: Options): Promise<Array<Document>>;
}
