"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable
const node_fetch_1 = require("node-fetch");
const abort_controller_1 = require("abort-controller");
const common_1 = require("./common");
exports.ValidationError = common_1.ValidationError;
const interfaces_1 = require("./interfaces");
class RequestError extends Error {
    constructor(message, 
    /**
     * The original error causing this request to fail
     * Inherits Error in case of network or parse errors
     * In case of an invalid HTTP response it will contain an object with the body/trimmed text of the response
     */
    cause, method, options) {
        super(message);
        this.cause = cause;
        this.method = method;
        this.options = options;
        this.name = 'RequestError';
    }
}
exports.RequestError = RequestError;
class TimeoutError extends Error {
    constructor(message, method, options) {
        super(message);
        this.method = method;
        this.options = options;
        this.name = 'TimeoutError';
    }
}
exports.TimeoutError = TimeoutError;
class DBClient {
    constructor(serverUrl, options = {}) {
        this.serverUrl = serverUrl;
        this.options = options;
        this.props = interfaces_1.schema.definitions.DB.properties;
        this.validators = DBClient.validators;
    }
    async get(ctx, key, options) {
        const body = {
            ctx,
            key,
        };
        const mergedOptions = Object.assign(Object.assign({ serverUrl: this.serverUrl }, this.options), options);
        const { fetchImplementation, timeoutMs, headers, serverUrl } = mergedOptions, fetchOptions = __rest(mergedOptions, ["fetchImplementation", "timeoutMs", "headers", "serverUrl"]);
        const fetchImpl = fetchImplementation || node_fetch_1.default;
        let timeout;
        if (timeoutMs) {
            const controller = new abort_controller_1.default();
            timeout = setTimeout(() => controller.abort(), timeoutMs);
            fetchOptions.signal = controller.signal;
        }
        let response;
        let responseBody;
        let responseText;
        let isJSON;
        try {
            response = await fetchImpl(`${serverUrl}/get`, Object.assign(Object.assign({}, fetchOptions), { headers: Object.assign(Object.assign({}, headers), { 'Content-Type': 'application/json' }), body: JSON.stringify(body), method: 'POST' }));
            isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
            if (isJSON) {
                responseBody = await response.json();
            }
            else {
                responseText = await response.text();
            }
        }
        catch (err) {
            if (err.message === 'The user aborted a request.') {
                timeout = undefined;
                throw new TimeoutError('Request aborted due to timeout', 'get', mergedOptions);
            }
            throw new RequestError(err.message, err, 'get', mergedOptions);
        }
        finally {
            if (timeout)
                clearTimeout(timeout);
        }
        if (response.status >= 200 && response.status < 300) {
            const validator = this.validators.get;
            const wrapped = { returns: responseBody }; // wrapped for coersion
            if (!validator(wrapped)) {
                throw new common_1.ValidationError('Failed to validate response', validator.errors);
            }
            return wrapped.returns;
        }
        else if (!isJSON) {
            // fall through to throw
        }
        else if (response.status === 400) {
            if (responseBody.name === 'ValidationError') {
                throw new common_1.ValidationError(responseBody.message, responseBody.errors);
            }
        }
        else if (response.status === 500) {
            throw new interfaces_1.InternalServerError(responseBody.message);
        }
        throw new RequestError(`${response.status} - ${response.statusText}`, { responseText: responseText && responseText.slice(0, 256), responseBody }, 'get', mergedOptions);
    }
    async getWithVersion(ctx, key, options) {
        const body = {
            ctx,
            key,
        };
        const mergedOptions = Object.assign(Object.assign({ serverUrl: this.serverUrl }, this.options), options);
        const { fetchImplementation, timeoutMs, headers, serverUrl } = mergedOptions, fetchOptions = __rest(mergedOptions, ["fetchImplementation", "timeoutMs", "headers", "serverUrl"]);
        const fetchImpl = fetchImplementation || node_fetch_1.default;
        let timeout;
        if (timeoutMs) {
            const controller = new abort_controller_1.default();
            timeout = setTimeout(() => controller.abort(), timeoutMs);
            fetchOptions.signal = controller.signal;
        }
        let response;
        let responseBody;
        let responseText;
        let isJSON;
        try {
            response = await fetchImpl(`${serverUrl}/getWithVersion`, Object.assign(Object.assign({}, fetchOptions), { headers: Object.assign(Object.assign({}, headers), { 'Content-Type': 'application/json' }), body: JSON.stringify(body), method: 'POST' }));
            isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
            if (isJSON) {
                responseBody = await response.json();
            }
            else {
                responseText = await response.text();
            }
        }
        catch (err) {
            if (err.message === 'The user aborted a request.') {
                timeout = undefined;
                throw new TimeoutError('Request aborted due to timeout', 'getWithVersion', mergedOptions);
            }
            throw new RequestError(err.message, err, 'getWithVersion', mergedOptions);
        }
        finally {
            if (timeout)
                clearTimeout(timeout);
        }
        if (response.status >= 200 && response.status < 300) {
            const validator = this.validators.getWithVersion;
            const wrapped = { returns: responseBody }; // wrapped for coersion
            if (!validator(wrapped)) {
                throw new common_1.ValidationError('Failed to validate response', validator.errors);
            }
            return wrapped.returns;
        }
        else if (!isJSON) {
            // fall through to throw
        }
        else if (response.status === 400) {
            if (responseBody.name === 'ValidationError') {
                throw new common_1.ValidationError(responseBody.message, responseBody.errors);
            }
        }
        else if (response.status === 500) {
            throw new interfaces_1.InternalServerError(responseBody.message);
        }
        throw new RequestError(`${response.status} - ${response.statusText}`, { responseText: responseText && responseText.slice(0, 256), responseBody }, 'getWithVersion', mergedOptions);
    }
    async setIfVersion(ctx, key, version, value, opts, options) {
        const body = {
            ctx,
            key,
            version,
            value,
            opts,
        };
        const mergedOptions = Object.assign(Object.assign({ serverUrl: this.serverUrl }, this.options), options);
        const { fetchImplementation, timeoutMs, headers, serverUrl } = mergedOptions, fetchOptions = __rest(mergedOptions, ["fetchImplementation", "timeoutMs", "headers", "serverUrl"]);
        const fetchImpl = fetchImplementation || node_fetch_1.default;
        let timeout;
        if (timeoutMs) {
            const controller = new abort_controller_1.default();
            timeout = setTimeout(() => controller.abort(), timeoutMs);
            fetchOptions.signal = controller.signal;
        }
        let response;
        let responseBody;
        let responseText;
        let isJSON;
        try {
            response = await fetchImpl(`${serverUrl}/setIfVersion`, Object.assign(Object.assign({}, fetchOptions), { headers: Object.assign(Object.assign({}, headers), { 'Content-Type': 'application/json' }), body: JSON.stringify(body), method: 'POST' }));
            isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
            if (isJSON) {
                responseBody = await response.json();
            }
            else {
                responseText = await response.text();
            }
        }
        catch (err) {
            if (err.message === 'The user aborted a request.') {
                timeout = undefined;
                throw new TimeoutError('Request aborted due to timeout', 'setIfVersion', mergedOptions);
            }
            throw new RequestError(err.message, err, 'setIfVersion', mergedOptions);
        }
        finally {
            if (timeout)
                clearTimeout(timeout);
        }
        if (response.status >= 200 && response.status < 300) {
            const validator = this.validators.setIfVersion;
            const wrapped = { returns: responseBody }; // wrapped for coersion
            if (!validator(wrapped)) {
                throw new common_1.ValidationError('Failed to validate response', validator.errors);
            }
            return wrapped.returns;
        }
        else if (!isJSON) {
            // fall through to throw
        }
        else if (response.status === 400) {
            if (responseBody.name === 'ValidationError') {
                throw new common_1.ValidationError(responseBody.message, responseBody.errors);
            }
        }
        else if (response.status === 500) {
            throw new interfaces_1.InternalServerError(responseBody.message);
        }
        throw new RequestError(`${response.status} - ${response.statusText}`, { responseText: responseText && responseText.slice(0, 256), responseBody }, 'setIfVersion', mergedOptions);
    }
    async startPolling(ctx, key, options) {
        const body = {
            ctx,
            key,
        };
        const mergedOptions = Object.assign(Object.assign({ serverUrl: this.serverUrl }, this.options), options);
        const { fetchImplementation, timeoutMs, headers, serverUrl } = mergedOptions, fetchOptions = __rest(mergedOptions, ["fetchImplementation", "timeoutMs", "headers", "serverUrl"]);
        const fetchImpl = fetchImplementation || node_fetch_1.default;
        let timeout;
        if (timeoutMs) {
            const controller = new abort_controller_1.default();
            timeout = setTimeout(() => controller.abort(), timeoutMs);
            fetchOptions.signal = controller.signal;
        }
        let response;
        let responseBody;
        let responseText;
        let isJSON;
        try {
            response = await fetchImpl(`${serverUrl}/startPolling`, Object.assign(Object.assign({}, fetchOptions), { headers: Object.assign(Object.assign({}, headers), { 'Content-Type': 'application/json' }), body: JSON.stringify(body), method: 'POST' }));
            isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
            if (isJSON) {
                responseBody = await response.json();
            }
            else {
                responseText = await response.text();
            }
        }
        catch (err) {
            if (err.message === 'The user aborted a request.') {
                timeout = undefined;
                throw new TimeoutError('Request aborted due to timeout', 'startPolling', mergedOptions);
            }
            throw new RequestError(err.message, err, 'startPolling', mergedOptions);
        }
        finally {
            if (timeout)
                clearTimeout(timeout);
        }
        if (response.status >= 200 && response.status < 300) {
            const validator = this.validators.startPolling;
            const wrapped = { returns: responseBody }; // wrapped for coersion
            if (!validator(wrapped)) {
                throw new common_1.ValidationError('Failed to validate response', validator.errors);
            }
            return wrapped.returns;
        }
        else if (!isJSON) {
            // fall through to throw
        }
        else if (response.status === 400) {
            if (responseBody.name === 'ValidationError') {
                throw new common_1.ValidationError(responseBody.message, responseBody.errors);
            }
        }
        else if (response.status === 500) {
            throw new interfaces_1.InternalServerError(responseBody.message);
        }
        throw new RequestError(`${response.status} - ${response.statusText}`, { responseText: responseText && responseText.slice(0, 256), responseBody }, 'startPolling', mergedOptions);
    }
    async poll(ctx, keysToVersions, opts, options) {
        const body = {
            ctx,
            keysToVersions,
            opts,
        };
        const mergedOptions = Object.assign(Object.assign({ serverUrl: this.serverUrl }, this.options), options);
        const { fetchImplementation, timeoutMs, headers, serverUrl } = mergedOptions, fetchOptions = __rest(mergedOptions, ["fetchImplementation", "timeoutMs", "headers", "serverUrl"]);
        const fetchImpl = fetchImplementation || node_fetch_1.default;
        let timeout;
        if (timeoutMs) {
            const controller = new abort_controller_1.default();
            timeout = setTimeout(() => controller.abort(), timeoutMs);
            fetchOptions.signal = controller.signal;
        }
        let response;
        let responseBody;
        let responseText;
        let isJSON;
        try {
            response = await fetchImpl(`${serverUrl}/poll`, Object.assign(Object.assign({}, fetchOptions), { headers: Object.assign(Object.assign({}, headers), { 'Content-Type': 'application/json' }), body: JSON.stringify(body), method: 'POST' }));
            isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
            if (isJSON) {
                responseBody = await response.json();
            }
            else {
                responseText = await response.text();
            }
        }
        catch (err) {
            if (err.message === 'The user aborted a request.') {
                timeout = undefined;
                throw new TimeoutError('Request aborted due to timeout', 'poll', mergedOptions);
            }
            throw new RequestError(err.message, err, 'poll', mergedOptions);
        }
        finally {
            if (timeout)
                clearTimeout(timeout);
        }
        if (response.status >= 200 && response.status < 300) {
            const validator = this.validators.poll;
            const wrapped = { returns: responseBody }; // wrapped for coersion
            if (!validator(wrapped)) {
                throw new common_1.ValidationError('Failed to validate response', validator.errors);
            }
            return wrapped.returns;
        }
        else if (!isJSON) {
            // fall through to throw
        }
        else if (response.status === 400) {
            if (responseBody.name === 'ValidationError') {
                throw new common_1.ValidationError(responseBody.message, responseBody.errors);
            }
        }
        else if (response.status === 500) {
            throw new interfaces_1.InternalServerError(responseBody.message);
        }
        throw new RequestError(`${response.status} - ${response.statusText}`, { responseText: responseText && responseText.slice(0, 256), responseBody }, 'poll', mergedOptions);
    }
    async create(ctx, key, value, options) {
        const body = {
            ctx,
            key,
            value,
        };
        const mergedOptions = Object.assign(Object.assign({ serverUrl: this.serverUrl }, this.options), options);
        const { fetchImplementation, timeoutMs, headers, serverUrl } = mergedOptions, fetchOptions = __rest(mergedOptions, ["fetchImplementation", "timeoutMs", "headers", "serverUrl"]);
        const fetchImpl = fetchImplementation || node_fetch_1.default;
        let timeout;
        if (timeoutMs) {
            const controller = new abort_controller_1.default();
            timeout = setTimeout(() => controller.abort(), timeoutMs);
            fetchOptions.signal = controller.signal;
        }
        let response;
        let responseBody;
        let responseText;
        let isJSON;
        try {
            response = await fetchImpl(`${serverUrl}/create`, Object.assign(Object.assign({}, fetchOptions), { headers: Object.assign(Object.assign({}, headers), { 'Content-Type': 'application/json' }), body: JSON.stringify(body), method: 'POST' }));
            isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
            if (isJSON) {
                responseBody = await response.json();
            }
            else {
                responseText = await response.text();
            }
        }
        catch (err) {
            if (err.message === 'The user aborted a request.') {
                timeout = undefined;
                throw new TimeoutError('Request aborted due to timeout', 'create', mergedOptions);
            }
            throw new RequestError(err.message, err, 'create', mergedOptions);
        }
        finally {
            if (timeout)
                clearTimeout(timeout);
        }
        if (response.status >= 200 && response.status < 300) {
            const validator = this.validators.create;
            const wrapped = { returns: responseBody }; // wrapped for coersion
            if (!validator(wrapped)) {
                throw new common_1.ValidationError('Failed to validate response', validator.errors);
            }
            return wrapped.returns;
        }
        else if (!isJSON) {
            // fall through to throw
        }
        else if (response.status === 400) {
            if (responseBody.name === 'ValidationError') {
                throw new common_1.ValidationError(responseBody.message, responseBody.errors);
            }
        }
        else if (response.status === 500) {
            throw new interfaces_1.InternalServerError(responseBody.message);
        }
        throw new RequestError(`${response.status} - ${response.statusText}`, { responseText: responseText && responseText.slice(0, 256), responseBody }, 'create', mergedOptions);
    }
    async remove(ctx, key, options) {
        const body = {
            ctx,
            key,
        };
        const mergedOptions = Object.assign(Object.assign({ serverUrl: this.serverUrl }, this.options), options);
        const { fetchImplementation, timeoutMs, headers, serverUrl } = mergedOptions, fetchOptions = __rest(mergedOptions, ["fetchImplementation", "timeoutMs", "headers", "serverUrl"]);
        const fetchImpl = fetchImplementation || node_fetch_1.default;
        let timeout;
        if (timeoutMs) {
            const controller = new abort_controller_1.default();
            timeout = setTimeout(() => controller.abort(), timeoutMs);
            fetchOptions.signal = controller.signal;
        }
        let response;
        let responseBody;
        let responseText;
        let isJSON;
        try {
            response = await fetchImpl(`${serverUrl}/remove`, Object.assign(Object.assign({}, fetchOptions), { headers: Object.assign(Object.assign({}, headers), { 'Content-Type': 'application/json' }), body: JSON.stringify(body), method: 'POST' }));
            isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
            if (isJSON) {
                responseBody = await response.json();
            }
            else {
                responseText = await response.text();
            }
        }
        catch (err) {
            if (err.message === 'The user aborted a request.') {
                timeout = undefined;
                throw new TimeoutError('Request aborted due to timeout', 'remove', mergedOptions);
            }
            throw new RequestError(err.message, err, 'remove', mergedOptions);
        }
        finally {
            if (timeout)
                clearTimeout(timeout);
        }
        if (response.status >= 200 && response.status < 300) {
            const validator = this.validators.remove;
            const wrapped = { returns: responseBody }; // wrapped for coersion
            if (!validator(wrapped)) {
                throw new common_1.ValidationError('Failed to validate response', validator.errors);
            }
            return wrapped.returns;
        }
        else if (!isJSON) {
            // fall through to throw
        }
        else if (response.status === 400) {
            if (responseBody.name === 'ValidationError') {
                throw new common_1.ValidationError(responseBody.message, responseBody.errors);
            }
        }
        else if (response.status === 500) {
            throw new interfaces_1.InternalServerError(responseBody.message);
        }
        throw new RequestError(`${response.status} - ${response.statusText}`, { responseText: responseText && responseText.slice(0, 256), responseBody }, 'remove', mergedOptions);
    }
    async find(ctx, query, options) {
        const body = {
            ctx,
            query,
        };
        const mergedOptions = Object.assign(Object.assign({ serverUrl: this.serverUrl }, this.options), options);
        const { fetchImplementation, timeoutMs, headers, serverUrl } = mergedOptions, fetchOptions = __rest(mergedOptions, ["fetchImplementation", "timeoutMs", "headers", "serverUrl"]);
        const fetchImpl = fetchImplementation || node_fetch_1.default;
        let timeout;
        if (timeoutMs) {
            const controller = new abort_controller_1.default();
            timeout = setTimeout(() => controller.abort(), timeoutMs);
            fetchOptions.signal = controller.signal;
        }
        let response;
        let responseBody;
        let responseText;
        let isJSON;
        try {
            response = await fetchImpl(`${serverUrl}/find`, Object.assign(Object.assign({}, fetchOptions), { headers: Object.assign(Object.assign({}, headers), { 'Content-Type': 'application/json' }), body: JSON.stringify(body), method: 'POST' }));
            isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
            if (isJSON) {
                responseBody = await response.json();
            }
            else {
                responseText = await response.text();
            }
        }
        catch (err) {
            if (err.message === 'The user aborted a request.') {
                timeout = undefined;
                throw new TimeoutError('Request aborted due to timeout', 'find', mergedOptions);
            }
            throw new RequestError(err.message, err, 'find', mergedOptions);
        }
        finally {
            if (timeout)
                clearTimeout(timeout);
        }
        if (response.status >= 200 && response.status < 300) {
            const validator = this.validators.find;
            const wrapped = { returns: responseBody }; // wrapped for coersion
            if (!validator(wrapped)) {
                throw new common_1.ValidationError('Failed to validate response', validator.errors);
            }
            return wrapped.returns;
        }
        else if (!isJSON) {
            // fall through to throw
        }
        else if (response.status === 400) {
            if (responseBody.name === 'ValidationError') {
                throw new common_1.ValidationError(responseBody.message, responseBody.errors);
            }
        }
        else if (response.status === 500) {
            throw new interfaces_1.InternalServerError(responseBody.message);
        }
        throw new RequestError(`${response.status} - ${response.statusText}`, { responseText: responseText && responseText.slice(0, 256), responseBody }, 'find', mergedOptions);
    }
}
exports.DBClient = DBClient;
DBClient.methods = [
    'get',
    'getWithVersion',
    'setIfVersion',
    'startPolling',
    'poll',
    'create',
    'remove',
    'find',
];
DBClient.validators = common_1.createReturnTypeValidator(interfaces_1.schema, 'DB');
//# sourceMappingURL=client.js.map