"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
const AjvCtor = require("ajv");
class ValidationError extends Error {
    constructor(message, errors) {
        super(message);
        this.errors = errors;
        this.name = 'ValidationError';
    }
}
exports.ValidationError = ValidationError;
function createValidator() {
    const ajv = new AjvCtor({ useDefaults: true, allErrors: true });
    ajv.addMetaSchema(require('ajv/lib/refs/json-schema-draft-06.json'));
    ajv.addKeyword('coerce-date', {
        type: 'string',
        modifying: true,
        valid: true,
        compile: (onOrOff, parentSchema) => {
            if (parentSchema.format !== 'date-time') {
                throw new Error('Format should be date-time when using coerce-date');
            }
            if (onOrOff !== true) {
                return lodash_1.identity;
            }
            return (v, _dataPath, obj, key) => {
                if (obj === undefined || key === undefined) {
                    throw new Error('Cannot coerce a date at root level');
                }
                obj[key] = new Date(v);
                return true;
            };
        },
    });
    return ajv;
}
// tslint:disable:max-line-length
/**
 * Copied from https://github.com/WHenderson/json-pointer-rfc6901/blob/ca2bfd17abe37ff09394222128192023dbfb067b/src/json-pointer.coffee#L255
 */
function escapeJsonPointer(ptr) {
    return encodeURIComponent(ptr.replace(/~/g, '~0').replace(/\//g, '~1'));
}
// tslint:enable:max-line-length
function createClassValidator(schema, className, field) {
    const ajv = createValidator();
    for (const [k, v] of Object.entries(schema.definitions)) {
        ajv.addSchema(v, `#/definitions/${escapeJsonPointer(k)}`);
    }
    return lodash_1.fromPairs(Object.entries(schema.definitions[className].properties).map(([method, s]) => [
        method, ajv.compile(s.properties[field]),
    ]));
}
exports.createClassValidator = createClassValidator;
function createReturnTypeValidator(schema, className) {
    const ajv = createValidator();
    for (const [k, v] of Object.entries(schema.definitions)) {
        ajv.addSchema(v, `#/definitions/${escapeJsonPointer(k)}`);
    }
    return lodash_1.fromPairs(Object.entries(schema.definitions[className].properties).map(([method, s]) => [
        method, ajv.compile({ properties: lodash_1.pick(s.properties, 'returns') }),
    ]));
}
exports.createReturnTypeValidator = createReturnTypeValidator;
//# sourceMappingURL=common.js.map