// tslint:disable
import fetch from 'node-fetch';
import { RequestInit, Response } from 'node-fetch';
import AbortController from 'abort-controller';
import { createReturnTypeValidator, ClassValidator, ValidationError } from './common';
import {
  schema,
  InternalServerError,
  Document,
  PathFilter,
  EqFilter,
  NeFilter,
  ComparableFilter,
  GtFilter,
  GteFilter,
  LtFilter,
  LteFilter,
  ExistsFilter,
  IsNullFilter,
  MatchesFilter,
  StartsWithFilter,
  AndFilter,
  OrFilter,
  NotFilter,
  Order,
  Query,
  ClientContext,
  ServerOnlyContext,
  UpdateOptions,
  Version,
  Patch,
  VersionedObject,
  VersionedMaybeObject,
  Patches,
  StoredDocument,
  PollOptions,
  Tombstone,
  AddOperation_of_any_end,
  RemoveOperation,
  ReplaceOperation_of_any_end,
  MoveOperation,
  CopyOperation,
  TestOperation_of_any_end,
  GetOperation_of_any_end,
  Direction,
  Serializable,
  Filter,
} from './interfaces';

export interface Options extends Pick<RequestInit, 'agent' | 'redirect' | 'follow' | 'compress'> {
  fetchImplementation?: typeof fetch;
  timeoutMs?: number;
  headers?: Record<string, string>;
}

export class RequestError extends Error {
  public readonly name = 'RequestError';
  constructor(
    message: string,
    /**
     * The original error causing this request to fail
     * Inherits Error in case of network or parse errors
     * In case of an invalid HTTP response it will contain an object with the body/trimmed text of the response
     */
    public readonly cause: any,
    public readonly method: string,
    public readonly options: any
  ) {
    super(message);
  }
}

export class TimeoutError extends Error {
  public readonly name = 'TimeoutError';
  constructor(message: string, public readonly method: string, public readonly options: any) {
    super(message);
  }
}

export {
  ValidationError,
};

export type Context = ClientContext;






























export interface DB {
  get(ctx: ClientContext ,key: string): Promise<any>;
  getWithVersion(ctx: ClientContext ,key: string): Promise<VersionedMaybeObject>;
  setIfVersion(ctx: ClientContext ,key: string, version: Version, value?: ({  }) | (Array<any>) | ((null) | (string) | (number) | (boolean)), opts?: UpdateOptions): Promise<boolean>;
  startPolling(ctx: ClientContext ,key: string): Promise<VersionedMaybeObject>;
  poll(ctx: ClientContext ,keysToVersions: Array<[string, Version]>, opts: PollOptions): Promise<Array<[string, Array<Patch>]>>;
  create(ctx: ClientContext ,key: string, value: ({  }) | (Array<any>) | ((null) | (string) | (number) | (boolean))): Promise<boolean>;
  remove(ctx: ClientContext ,key: string): Promise<boolean>;
  find(ctx: ClientContext ,query: Query): Promise<Array<Document>>;
}

export class DBClient {
  public static readonly methods = [
    'get',
    'getWithVersion',
    'setIfVersion',
    'startPolling',
    'poll',
    'create',
    'remove',
    'find',
  ];
  public static readonly validators: ClassValidator = createReturnTypeValidator(schema, 'DB');

  protected readonly props = schema.definitions.DB.properties;

  public readonly validators: ClassValidator; // We don't have class name in method scope because mustache sux

  public constructor(public readonly serverUrl: string, protected readonly options: Options = {}) {
    this.validators = DBClient.validators;
  }

  public async get(ctx: ClientContext ,key: string, options?: Options): Promise<any> {
    const body = {
      ctx,
      key,
    };

    const mergedOptions = {
      serverUrl: this.serverUrl,
      ...this.options,
      ...options,
    };

    const { fetchImplementation, timeoutMs, headers, serverUrl, ...fetchOptions } = mergedOptions;

    const fetchImpl = fetchImplementation || fetch;

    let timeout: NodeJS.Timeout | undefined;
    if (timeoutMs) {
      const controller = new AbortController();
      timeout = setTimeout(() => controller.abort(), timeoutMs);
      (fetchOptions as any).signal = controller.signal;
    }

    let response: Response;
    let responseBody: any;
    let responseText: string | undefined;
    let isJSON: boolean;
    try {
      response = await fetchImpl(`${serverUrl}/get`, {
        ...fetchOptions,
        headers: {
          ...headers,
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
        method: 'POST',
      });
      isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
      if (isJSON) {
        responseBody = await response.json();
      } else {
        responseText = await response.text();
      }
    } catch (err) {
      if (err.message === 'The user aborted a request.') {
        timeout = undefined;
        throw new TimeoutError('Request aborted due to timeout', 'get', mergedOptions);
      }
      throw new RequestError(err.message, err, 'get', mergedOptions);
    } finally {
      if (timeout) clearTimeout(timeout);
    }
    if (response.status >= 200 && response.status < 300) {
      const validator = this.validators.get;
      const wrapped = { returns: responseBody }; // wrapped for coersion
      if (!validator(wrapped)) {
        throw new ValidationError('Failed to validate response', validator.errors);
      }
      return wrapped.returns as any;
    } else if (!isJSON) {
      // fall through to throw
    } else if (response.status === 400) {
      if (responseBody.name === 'ValidationError') {
        throw new ValidationError(responseBody.message, responseBody.errors);
      }
    } else if (response.status === 500) {
      throw new InternalServerError(responseBody.message);
    }
    throw new RequestError(`${response.status} - ${response.statusText}`,
      { responseText: responseText && responseText.slice(0, 256), responseBody },
      'get',
      mergedOptions);
  }

  public async getWithVersion(ctx: ClientContext ,key: string, options?: Options): Promise<VersionedMaybeObject> {
    const body = {
      ctx,
      key,
    };

    const mergedOptions = {
      serverUrl: this.serverUrl,
      ...this.options,
      ...options,
    };

    const { fetchImplementation, timeoutMs, headers, serverUrl, ...fetchOptions } = mergedOptions;

    const fetchImpl = fetchImplementation || fetch;

    let timeout: NodeJS.Timeout | undefined;
    if (timeoutMs) {
      const controller = new AbortController();
      timeout = setTimeout(() => controller.abort(), timeoutMs);
      (fetchOptions as any).signal = controller.signal;
    }

    let response: Response;
    let responseBody: any;
    let responseText: string | undefined;
    let isJSON: boolean;
    try {
      response = await fetchImpl(`${serverUrl}/getWithVersion`, {
        ...fetchOptions,
        headers: {
          ...headers,
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
        method: 'POST',
      });
      isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
      if (isJSON) {
        responseBody = await response.json();
      } else {
        responseText = await response.text();
      }
    } catch (err) {
      if (err.message === 'The user aborted a request.') {
        timeout = undefined;
        throw new TimeoutError('Request aborted due to timeout', 'getWithVersion', mergedOptions);
      }
      throw new RequestError(err.message, err, 'getWithVersion', mergedOptions);
    } finally {
      if (timeout) clearTimeout(timeout);
    }
    if (response.status >= 200 && response.status < 300) {
      const validator = this.validators.getWithVersion;
      const wrapped = { returns: responseBody }; // wrapped for coersion
      if (!validator(wrapped)) {
        throw new ValidationError('Failed to validate response', validator.errors);
      }
      return wrapped.returns as VersionedMaybeObject;
    } else if (!isJSON) {
      // fall through to throw
    } else if (response.status === 400) {
      if (responseBody.name === 'ValidationError') {
        throw new ValidationError(responseBody.message, responseBody.errors);
      }
    } else if (response.status === 500) {
      throw new InternalServerError(responseBody.message);
    }
    throw new RequestError(`${response.status} - ${response.statusText}`,
      { responseText: responseText && responseText.slice(0, 256), responseBody },
      'getWithVersion',
      mergedOptions);
  }

  public async setIfVersion(ctx: ClientContext ,key: string, version: Version, value?: ({  }) | (Array<any>) | ((null) | (string) | (number) | (boolean)), opts?: UpdateOptions, options?: Options): Promise<boolean> {
    const body = {
      ctx,
      key,
      version,
      value,
      opts,
    };

    const mergedOptions = {
      serverUrl: this.serverUrl,
      ...this.options,
      ...options,
    };

    const { fetchImplementation, timeoutMs, headers, serverUrl, ...fetchOptions } = mergedOptions;

    const fetchImpl = fetchImplementation || fetch;

    let timeout: NodeJS.Timeout | undefined;
    if (timeoutMs) {
      const controller = new AbortController();
      timeout = setTimeout(() => controller.abort(), timeoutMs);
      (fetchOptions as any).signal = controller.signal;
    }

    let response: Response;
    let responseBody: any;
    let responseText: string | undefined;
    let isJSON: boolean;
    try {
      response = await fetchImpl(`${serverUrl}/setIfVersion`, {
        ...fetchOptions,
        headers: {
          ...headers,
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
        method: 'POST',
      });
      isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
      if (isJSON) {
        responseBody = await response.json();
      } else {
        responseText = await response.text();
      }
    } catch (err) {
      if (err.message === 'The user aborted a request.') {
        timeout = undefined;
        throw new TimeoutError('Request aborted due to timeout', 'setIfVersion', mergedOptions);
      }
      throw new RequestError(err.message, err, 'setIfVersion', mergedOptions);
    } finally {
      if (timeout) clearTimeout(timeout);
    }
    if (response.status >= 200 && response.status < 300) {
      const validator = this.validators.setIfVersion;
      const wrapped = { returns: responseBody }; // wrapped for coersion
      if (!validator(wrapped)) {
        throw new ValidationError('Failed to validate response', validator.errors);
      }
      return wrapped.returns as boolean;
    } else if (!isJSON) {
      // fall through to throw
    } else if (response.status === 400) {
      if (responseBody.name === 'ValidationError') {
        throw new ValidationError(responseBody.message, responseBody.errors);
      }
    } else if (response.status === 500) {
      throw new InternalServerError(responseBody.message);
    }
    throw new RequestError(`${response.status} - ${response.statusText}`,
      { responseText: responseText && responseText.slice(0, 256), responseBody },
      'setIfVersion',
      mergedOptions);
  }

  public async startPolling(ctx: ClientContext ,key: string, options?: Options): Promise<VersionedMaybeObject> {
    const body = {
      ctx,
      key,
    };

    const mergedOptions = {
      serverUrl: this.serverUrl,
      ...this.options,
      ...options,
    };

    const { fetchImplementation, timeoutMs, headers, serverUrl, ...fetchOptions } = mergedOptions;

    const fetchImpl = fetchImplementation || fetch;

    let timeout: NodeJS.Timeout | undefined;
    if (timeoutMs) {
      const controller = new AbortController();
      timeout = setTimeout(() => controller.abort(), timeoutMs);
      (fetchOptions as any).signal = controller.signal;
    }

    let response: Response;
    let responseBody: any;
    let responseText: string | undefined;
    let isJSON: boolean;
    try {
      response = await fetchImpl(`${serverUrl}/startPolling`, {
        ...fetchOptions,
        headers: {
          ...headers,
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
        method: 'POST',
      });
      isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
      if (isJSON) {
        responseBody = await response.json();
      } else {
        responseText = await response.text();
      }
    } catch (err) {
      if (err.message === 'The user aborted a request.') {
        timeout = undefined;
        throw new TimeoutError('Request aborted due to timeout', 'startPolling', mergedOptions);
      }
      throw new RequestError(err.message, err, 'startPolling', mergedOptions);
    } finally {
      if (timeout) clearTimeout(timeout);
    }
    if (response.status >= 200 && response.status < 300) {
      const validator = this.validators.startPolling;
      const wrapped = { returns: responseBody }; // wrapped for coersion
      if (!validator(wrapped)) {
        throw new ValidationError('Failed to validate response', validator.errors);
      }
      return wrapped.returns as VersionedMaybeObject;
    } else if (!isJSON) {
      // fall through to throw
    } else if (response.status === 400) {
      if (responseBody.name === 'ValidationError') {
        throw new ValidationError(responseBody.message, responseBody.errors);
      }
    } else if (response.status === 500) {
      throw new InternalServerError(responseBody.message);
    }
    throw new RequestError(`${response.status} - ${response.statusText}`,
      { responseText: responseText && responseText.slice(0, 256), responseBody },
      'startPolling',
      mergedOptions);
  }

  public async poll(ctx: ClientContext ,keysToVersions: Array<[string, Version]>, opts: PollOptions, options?: Options): Promise<Array<[string, Array<Patch>]>> {
    const body = {
      ctx,
      keysToVersions,
      opts,
    };

    const mergedOptions = {
      serverUrl: this.serverUrl,
      ...this.options,
      ...options,
    };

    const { fetchImplementation, timeoutMs, headers, serverUrl, ...fetchOptions } = mergedOptions;

    const fetchImpl = fetchImplementation || fetch;

    let timeout: NodeJS.Timeout | undefined;
    if (timeoutMs) {
      const controller = new AbortController();
      timeout = setTimeout(() => controller.abort(), timeoutMs);
      (fetchOptions as any).signal = controller.signal;
    }

    let response: Response;
    let responseBody: any;
    let responseText: string | undefined;
    let isJSON: boolean;
    try {
      response = await fetchImpl(`${serverUrl}/poll`, {
        ...fetchOptions,
        headers: {
          ...headers,
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
        method: 'POST',
      });
      isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
      if (isJSON) {
        responseBody = await response.json();
      } else {
        responseText = await response.text();
      }
    } catch (err) {
      if (err.message === 'The user aborted a request.') {
        timeout = undefined;
        throw new TimeoutError('Request aborted due to timeout', 'poll', mergedOptions);
      }
      throw new RequestError(err.message, err, 'poll', mergedOptions);
    } finally {
      if (timeout) clearTimeout(timeout);
    }
    if (response.status >= 200 && response.status < 300) {
      const validator = this.validators.poll;
      const wrapped = { returns: responseBody }; // wrapped for coersion
      if (!validator(wrapped)) {
        throw new ValidationError('Failed to validate response', validator.errors);
      }
      return wrapped.returns as Array<[string, Array<Patch>]>;
    } else if (!isJSON) {
      // fall through to throw
    } else if (response.status === 400) {
      if (responseBody.name === 'ValidationError') {
        throw new ValidationError(responseBody.message, responseBody.errors);
      }
    } else if (response.status === 500) {
      throw new InternalServerError(responseBody.message);
    }
    throw new RequestError(`${response.status} - ${response.statusText}`,
      { responseText: responseText && responseText.slice(0, 256), responseBody },
      'poll',
      mergedOptions);
  }

  public async create(ctx: ClientContext ,key: string, value: ({  }) | (Array<any>) | ((null) | (string) | (number) | (boolean)), options?: Options): Promise<boolean> {
    const body = {
      ctx,
      key,
      value,
    };

    const mergedOptions = {
      serverUrl: this.serverUrl,
      ...this.options,
      ...options,
    };

    const { fetchImplementation, timeoutMs, headers, serverUrl, ...fetchOptions } = mergedOptions;

    const fetchImpl = fetchImplementation || fetch;

    let timeout: NodeJS.Timeout | undefined;
    if (timeoutMs) {
      const controller = new AbortController();
      timeout = setTimeout(() => controller.abort(), timeoutMs);
      (fetchOptions as any).signal = controller.signal;
    }

    let response: Response;
    let responseBody: any;
    let responseText: string | undefined;
    let isJSON: boolean;
    try {
      response = await fetchImpl(`${serverUrl}/create`, {
        ...fetchOptions,
        headers: {
          ...headers,
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
        method: 'POST',
      });
      isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
      if (isJSON) {
        responseBody = await response.json();
      } else {
        responseText = await response.text();
      }
    } catch (err) {
      if (err.message === 'The user aborted a request.') {
        timeout = undefined;
        throw new TimeoutError('Request aborted due to timeout', 'create', mergedOptions);
      }
      throw new RequestError(err.message, err, 'create', mergedOptions);
    } finally {
      if (timeout) clearTimeout(timeout);
    }
    if (response.status >= 200 && response.status < 300) {
      const validator = this.validators.create;
      const wrapped = { returns: responseBody }; // wrapped for coersion
      if (!validator(wrapped)) {
        throw new ValidationError('Failed to validate response', validator.errors);
      }
      return wrapped.returns as boolean;
    } else if (!isJSON) {
      // fall through to throw
    } else if (response.status === 400) {
      if (responseBody.name === 'ValidationError') {
        throw new ValidationError(responseBody.message, responseBody.errors);
      }
    } else if (response.status === 500) {
      throw new InternalServerError(responseBody.message);
    }
    throw new RequestError(`${response.status} - ${response.statusText}`,
      { responseText: responseText && responseText.slice(0, 256), responseBody },
      'create',
      mergedOptions);
  }

  public async remove(ctx: ClientContext ,key: string, options?: Options): Promise<boolean> {
    const body = {
      ctx,
      key,
    };

    const mergedOptions = {
      serverUrl: this.serverUrl,
      ...this.options,
      ...options,
    };

    const { fetchImplementation, timeoutMs, headers, serverUrl, ...fetchOptions } = mergedOptions;

    const fetchImpl = fetchImplementation || fetch;

    let timeout: NodeJS.Timeout | undefined;
    if (timeoutMs) {
      const controller = new AbortController();
      timeout = setTimeout(() => controller.abort(), timeoutMs);
      (fetchOptions as any).signal = controller.signal;
    }

    let response: Response;
    let responseBody: any;
    let responseText: string | undefined;
    let isJSON: boolean;
    try {
      response = await fetchImpl(`${serverUrl}/remove`, {
        ...fetchOptions,
        headers: {
          ...headers,
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
        method: 'POST',
      });
      isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
      if (isJSON) {
        responseBody = await response.json();
      } else {
        responseText = await response.text();
      }
    } catch (err) {
      if (err.message === 'The user aborted a request.') {
        timeout = undefined;
        throw new TimeoutError('Request aborted due to timeout', 'remove', mergedOptions);
      }
      throw new RequestError(err.message, err, 'remove', mergedOptions);
    } finally {
      if (timeout) clearTimeout(timeout);
    }
    if (response.status >= 200 && response.status < 300) {
      const validator = this.validators.remove;
      const wrapped = { returns: responseBody }; // wrapped for coersion
      if (!validator(wrapped)) {
        throw new ValidationError('Failed to validate response', validator.errors);
      }
      return wrapped.returns as boolean;
    } else if (!isJSON) {
      // fall through to throw
    } else if (response.status === 400) {
      if (responseBody.name === 'ValidationError') {
        throw new ValidationError(responseBody.message, responseBody.errors);
      }
    } else if (response.status === 500) {
      throw new InternalServerError(responseBody.message);
    }
    throw new RequestError(`${response.status} - ${response.statusText}`,
      { responseText: responseText && responseText.slice(0, 256), responseBody },
      'remove',
      mergedOptions);
  }

  public async find(ctx: ClientContext ,query: Query, options?: Options): Promise<Array<Document>> {
    const body = {
      ctx,
      query,
    };

    const mergedOptions = {
      serverUrl: this.serverUrl,
      ...this.options,
      ...options,
    };

    const { fetchImplementation, timeoutMs, headers, serverUrl, ...fetchOptions } = mergedOptions;

    const fetchImpl = fetchImplementation || fetch;

    let timeout: NodeJS.Timeout | undefined;
    if (timeoutMs) {
      const controller = new AbortController();
      timeout = setTimeout(() => controller.abort(), timeoutMs);
      (fetchOptions as any).signal = controller.signal;
    }

    let response: Response;
    let responseBody: any;
    let responseText: string | undefined;
    let isJSON: boolean;
    try {
      response = await fetchImpl(`${serverUrl}/find`, {
        ...fetchOptions,
        headers: {
          ...headers,
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
        method: 'POST',
      });
      isJSON = (response.headers.get('content-type') || '').startsWith('application/json');
      if (isJSON) {
        responseBody = await response.json();
      } else {
        responseText = await response.text();
      }
    } catch (err) {
      if (err.message === 'The user aborted a request.') {
        timeout = undefined;
        throw new TimeoutError('Request aborted due to timeout', 'find', mergedOptions);
      }
      throw new RequestError(err.message, err, 'find', mergedOptions);
    } finally {
      if (timeout) clearTimeout(timeout);
    }
    if (response.status >= 200 && response.status < 300) {
      const validator = this.validators.find;
      const wrapped = { returns: responseBody }; // wrapped for coersion
      if (!validator(wrapped)) {
        throw new ValidationError('Failed to validate response', validator.errors);
      }
      return wrapped.returns as Array<Document>;
    } else if (!isJSON) {
      // fall through to throw
    } else if (response.status === 400) {
      if (responseBody.name === 'ValidationError') {
        throw new ValidationError(responseBody.message, responseBody.errors);
      }
    } else if (response.status === 500) {
      throw new InternalServerError(responseBody.message);
    }
    throw new RequestError(`${response.status} - ${response.statusText}`,
      { responseText: responseText && responseText.slice(0, 256), responseBody },
      'find',
      mergedOptions);
  }
}








